import { APIMessage, GatewayInteractionCreateDispatchData, InteractionType, MessageType, RESTPostAPIApplicationCommandsJSONBody, Snowflake } from 'discord-api-types'

import { CommandContext } from './CommandContext'

import { traverseObject } from '../utils/UtilityFunctions'

import { CommandOptions, CommandType, Worker, CommandContext as ctx } from '../typings/lib'
import Collection from '@discordjs/collection'

import fs from 'fs'
import path from 'path'
import { Interaction, SlashCommandContext } from './SlashCommandContext'
import Pieces from '../utils/Pieces'

type MiddlewareFunction = (ctx: ctx) => boolean | Promise<boolean>

/**
 * Error in command
 */
export class CommandError extends Error {
  nonFatal?: boolean
}

/**
 * Utility in charge of holding and running commands
 */
export class CommandHandler {
  private added: boolean = false
  private addedInteractions: boolean = false

  private _options: CommandHandlerOptions = {
    default: {},
    bots: false,
    mentionPrefix: true,
    caseInsensitivePrefix: true,
    caseInsensitiveCommand: true,
    reuseInteractions: true
  }

  public middlewares: MiddlewareFunction[] = []
  public commands?: Collection<CommandType, CommandOptions>

  public CommandContext = CommandContext
  public SlashCommandContext = SlashCommandContext

  /**
   * Create's new Command Handler
   * @param worker Worker
   */
  constructor (private readonly worker: Worker) {}

  async setupInteractions (): Promise<void> {
    this.addedInteractions = true

    if (this.commands) {
      const interactions = this.commands.filter(x => !!x.interaction)
      if (interactions.size > 0) {
        this.worker.on('INTERACTION_CREATE', (data) => {
          this._interactionExec(data).catch(() => {})
        })

        if (this.worker.comms.id !== '0') return

        if (this._options.reuseInteractions) {
          const currentInteractions = await this.worker.api.interactions.get(this.worker.user.id, this._options.interactionGuild)

          const newInteractions = interactions.filter(command => !currentInteractions.find(interaction => command.interaction?.name === interaction.name))
          const deletedInteractions = currentInteractions.filter(interaction => !interactions.find(command => interaction.name === command.interaction?.name))
          const changedInteractions = interactions.filter(command => {
            if (command.interaction) {
              command.interaction.default_permission = typeof command.interaction.default_permission === 'boolean' ? command.interaction.default_permission : true
              traverseObject(command.interaction, obj => {
                if (typeof obj.required === 'boolean' && !obj.required) delete obj.required
              })
            }

            return !currentInteractions.find(interaction => {
              const current = Pieces.generate(command.interaction?.options)
              const incoming = Pieces.generate(interaction?.options)

              return interaction.default_permission === command.interaction?.default_permission &&
              interaction.description === command.interaction?.description &&
              interaction.name === command.interaction?.name &&
              Object.keys(current).every(x => current[x] === incoming[x]) &&
              Object.keys(incoming).every(x => incoming[x] === current[x])
            }) && !newInteractions.find(newCommand => newCommand === command)
          })

          const promises: Array<Promise<any>> = []

          newInteractions.forEach(command => promises.push(this.worker.api.interactions.add(command.interaction as RESTPostAPIApplicationCommandsJSONBody, this.worker.user.id, this._options.interactionGuild)))
          deletedInteractions.forEach(interaction => promises.push(this.worker.api.interactions.delete(interaction.id, this.worker.user.id, this._options.interactionGuild)))
          changedInteractions.forEach(command => promises.push(this.worker.api.interactions.add(command.interaction as RESTPostAPIApplicationCommandsJSONBody, this.worker.user.id, this._options.interactionGuild)))

          Promise.all(promises)
            .then(() => {
              this.worker.log(`Added ${newInteractions.size}, deleted ${deletedInteractions.length}, and updated ${changedInteractions.size} command interactions`)
            })
            .catch(err => {
              err.message = `${err.message as string} (Whilst posting Command Interactions)`
              console.error(err)
            })
        } else {
          this.worker.api.interactions.set(interactions.map(x => x.interaction) as RESTPostAPIApplicationCommandsJSONBody[], this.worker.user.id, this._options.interactionGuild)
            .then(() => {
              this.worker.log('Posted command interactions')
            })
            .catch(err => {
              err.message = `${err.message as string} (Whilst posting Command Interactions)`
              console.error(err)
            })
        }
      }
    }
  }

  public prefixFunction?: ((message: APIMessage) => Promise<string|string[]> | string|string[])
  public errorFunction = (ctx: ctx, err: CommandError): void => {
    if (ctx.myPerms('sendMessages')) {
      if (ctx.isInteraction || ctx.myPerms('embed')) {
        ctx.embed
          .color(0xFF0000)
          .title('An Error Occured')
          .description(`\`\`\`xl\n${err.message}\`\`\``)
          .send(true, false, true).catch(console.error)
      } else {
        ctx
          .send(`An Error Occured\n\`\`\`xl\n${err.message}\`\`\``)
          .catch(() => { })
      }
    }

    if (err.nonFatal) return

    err.message += ` (While Running Command: ${String(ctx.command.command)})`
    console.error(err)
  }

  /**
   * Load a directory of CommandOptions commands (will also load sub-folders)
   * @param directory Absolute directory full of command files
   */
  load (directory: string): this {
    if (!path.isAbsolute(directory)) directory = path.resolve(process.cwd(), directory)

    const files = fs.readdirSync(directory, { withFileTypes: true })

    files.forEach(file => {
      if (file.isDirectory()) return this.load(path.resolve(directory, file.name))

      if (!file.name.endsWith('.js')) return

      // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
      delete require.cache[require.resolve(path.resolve(directory, file.name))]

      // eslint-disable-next-line @typescript-eslint/no-var-requires
      let command = require(path.resolve(directory, file.name))

      if (!command) return

      if (command.default) command = command.default

      this.add(command)
    })

    return this
  }

  /**
   * Sets Command Handler options
   * @param opts Options
   * @returns this
   */
  options (opts: CommandHandlerOptions): this {
    this._options = {
      ...this._options,
      ...opts
    }

    return this
  }

  /**
   * Sets a prefix
   * @param fn String of prefix or Function to choose prefix with
   * @example
   * worker.commands
   *   .prefix('!')
   * // or
   *   .prefix(['!', '+'])
   * // or
   *   .prefix((message) => {
   *     return db.getPrefix(message.guild_id)
   *   })
   * @returns this
   */
  prefix (fn: string|string[] | ((message: APIMessage) => Promise<string|string[]> | string|string[])): this {
    if (Array.isArray(fn) || typeof fn === 'string') {
      this.prefixFunction = () => fn
    } else {
      this.prefixFunction = fn
    }

    return this
  }

  /**
   * Sets a prefix
   * @deprecated Use the new {@link CommandHandler.prefix | .prefix()} function
   * @param fn String of prefix or Function to choose prefix with
   * @returns this
   */
  public setPrefix (fn: string|string[] | ((message: APIMessage) => Promise<string|string[]> | string|string[])): this {
    console.warn('.setPrefix is deprecated, please use .prefix() instead.')

    return this.prefix(fn)
  }

  /**
   * Defines an error handler replacing the default one
   * @param fn Function to handle error
   * @example
   * worker.commands
   *  .error((ctx, error) => {
   *    ctx.send(`Error: ${error.message}`)
   *  })
   * @returns this
   */
  error (fn: (ctx: ctx, error: CommandError) => void): this {
    this.errorFunction = fn

    return this
  }

  /**
   * Adds a global middleware function
   * @param fn Middleware function
   * @returns this
   */
  middleware (fn: MiddlewareFunction): this {
    this.middlewares.push(fn)

    return this
  }

  /**
   * Adds a command to the command handler
   * @param command Command data, be sure to add exec() and command:
   * @example
   * worker.commands
   *   .add({
   *     command: 'hello',
   *     exec: (ctx) => {
   *       ctx.reply('World!')
   *     }
   *   })
   * @returns this
   */
  add (command: CommandOptions): this {
    if (!this.added) {
      this.added = true
      this.commands = new Collection()

      this.worker.on('MESSAGE_CREATE', (data) => {
        this._exec(data).catch(() => {})
      })

      this.worker.once('READY', () => {
        void this.setupInteractions()
      })
    }
    if (this.worker.comms.id === '0' && this.addedInteractions && command.interaction) {
      this.worker.api.interactions.add(command.interaction, this.worker.user.id, this._options.interactionGuild)
        .catch(err => {
          err.message = `${err.message as string} (Whilst posting a Command Interaction)`
          console.error(err)
        })
    }

    this.commands?.set(command.command, {
      ...this._options.default,
      ...command
    })

    return this
  }

  private _test (command: string, cmd: CommandType): boolean {
    if (this._options.caseInsensitiveCommand) command = command.toLowerCase()
    if (typeof cmd === 'string') return command === cmd
    if (cmd instanceof RegExp) return !!command.match(cmd)

    return false
  }

  /**
   * Gets a command from registry
   * @param command Command name to fetch
   * @param interaction Whether or not to look for interactions
   * @returns Command
   */
  public find (command: string, interaction?: boolean): CommandOptions | undefined {
    if (interaction) {
      return this.commands?.find(x => !!x.interaction && x.interaction.name === command)
    }

    return this.commands?.find(x => (this._test(command, x.command) || x.aliases?.some(alias => this._test(command, alias)) as boolean))
  }

  /**
   * Gets a command from registry
   * @deprecated Use the new {@link CommandHandler.find | .find()} function
   * @param command Command name to fetch
   * @returns Command
   */
  public findCommand (command: string): CommandOptions | undefined {
    console.warn('.findCommand is deprecated, please use .find() instead.')

    return this.find(command)
  }

  private async _interactionExec (data: GatewayInteractionCreateDispatchData): Promise<void> {
    if (!data.member) return
    if (data.type === InteractionType.MessageComponent) return

    const cmd = this.find(data.data.name, true) as CommandOptions
    if (!cmd) return

    const ctx = new this.SlashCommandContext({
      worker: this.worker,
      interaction: data as Interaction,
      command: cmd,
      prefix: '/',
      ran: data.data.name,
      args: data.data.options?.map(x => 'value' in x ? x.value : '').filter(x => x) ?? []
    })

    try {
      for (const midFn of this.middlewares) {
        try {
          // eslint-disable-next-line @typescript-eslint/no-unnecessary-boolean-literal-compare
          if (await midFn(ctx as unknown as CommandContext) !== true) return
        } catch (err) {
          err.nonFatal = true

          throw err
        }
      }
      await cmd.exec(ctx as unknown as CommandContext)
    } catch (err) {
      this.errorFunction(ctx as unknown as CommandContext, err)
    }
  }

  private async _exec (data: APIMessage): Promise<void> {
    if (!data.content || (!this._options.bots && data.author.bot)) return
    if (![MessageType.Default, MessageType.Reply].includes(data.type)) return

    let prefix: string | string[] | undefined = ''
    if (this.prefixFunction) {
      prefix = await this.prefixFunction(data)
      if (!Array.isArray(prefix)) prefix = [prefix]

      if (this._options.mentionPrefix) prefix.push(`<@${this.worker.user.id}>`, `<@!${this.worker.user.id}>`)

      const content = this._options.caseInsensitivePrefix ? data.content.toLowerCase() : data.content

      prefix = prefix.find(x => content.startsWith(x))
      if (!prefix) return
    }

    const args = data.content.slice(prefix ? prefix.length : 0).split(/\s/)
    if (args[0] === '') {
      args.shift()

      prefix += ' '
    }

    const command = args.shift() ?? ''

    const cmd = this.find(command)
    if (!cmd) return

    const ctx = new this.CommandContext({
      worker: this.worker,
      message: data,
      command: cmd,
      prefix,
      ran: command,
      args: args
    })

    try {
      for (const midFn of this.middlewares) {
        try {
          // eslint-disable-next-line @typescript-eslint/no-unnecessary-boolean-literal-compare
          if (await midFn(ctx) !== true) return
        } catch (err) {
          err.nonFatal = true

          throw err
        }
      }
      await cmd.exec(ctx)
    } catch (err) {
      this.errorFunction(ctx, err)
    }
  }
}

export interface CommandHandlerOptions {
  /**
   * Default CommandOptions ('command', 'exec', and 'aliases' cannot be defaulted)
   */
  default?: Partial<Pick<CommandOptions, Exclude<keyof CommandOptions, 'command' | 'exec' | 'aliases'>>>
  /**
   * Allow commands from bots
   * @default false
   */
  bots?: boolean
  /**
   * Whether or not to respond to your bot's @Mention
   * @default true
   */
  mentionPrefix?: boolean
  /**
   * Whether or not the prefix is case insensitive
   * @default true
   */
  caseInsensitivePrefix?: boolean
  /**
   * Whether or not the command is case insensitive
   * @default true
   */
  caseInsensitiveCommand?: boolean
  /**
   * Only post interaction to one specific guild (ID)
   */
  interactionGuild?: Snowflake
  /**
   * If interactions previously posted should be reused when possible
   * @default true
   */
  reuseInteractions?: boolean
}
